# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from maas.client.bones import SessionAPI

from bmcmanager.logs import log
from bmcmanager.dcim.base import DcimBase, DcimError


class MaaS(DcimBase):
    def __init__(self, args, config):
        super(MaaS, self).__init__(args, config)
        self._session = None

        if not self.dcim_params['api_url']:
            raise DcimError('MaaS API URL is not set, see README.md')
        if not self.dcim_params['api_key']:
            raise DcimError('MaaS API Key is not set, see README.md')

        self.api_url = self.dcim_params['api_url']
        self.api_key = self.dcim_params['api_key']

    def session(self) -> SessionAPI:
        """
        Return a sessions to the MaaS server. Subsequent calls reuse the same
        connection
        """
        if self._session:
            return self._session

        log.info(f'Connecting to {self.api_url}')
        _, self._session = SessionAPI.connect(self.api_url, apikey=self.api_key)

        return self._session

    def get_oobs(self):
        for machine in self.session().Machines.read(hostname=[self.identifier]):
            power = self.session().Machine.power_parameters(system_id=machine['system_id'])
            yield {
                'asset_tag': '',
                'ipmi': power['power_address'],
                'oob': machine['hardware_info']['mainboard_vendor'].lower(),
                'info': {
                    'id': machine['system_id'],
                    'name': machine['hostname'],
                    'display_name': machine['hostname'],
                    'serial': machine['hardware_info']['system_serial'],
                    'ipmi': power['power_address'],
                    'manufacturer': machine['hardware_info']['mainboard_vendor'],
                    'device_type': machine['hardware_info']['mainboard_product'],
                    'bios': machine['owner_data'].get('BIOS', ''),
                    'tsm': machine['owner_data'].get('TSM', ''),
                    'psu': machine['owner_data'].get('PSU', ''),
                    'site': machine['domain']['name'],
                    'status': machine['status_name'],
                },
                'identifier': machine['hostname'],
                'custom_fields': machine['owner_data'],
            }

    def get_secret(self, role, oob_info):
        power = self.session().Machine.power_parameters(system_id=oob_info['info']['id'])
        return {
            'name': power['power_user'],
            'plaintext': power['power_pass'],
        }

    def set_custom_fields(self, oob_info, custom_fields):
        raise DcimError('not support by MaaS DCIM')

    def oob_url(self, oob_info):
        if not self.dcim_params.get('ui_url'):
            raise DcimError('MaaS UI URL is not set, see README.md')
        return '{}/MAAS/l/machine/{}'.format(
            self.dcim_params['ui_url'], oob_info['info']['id'])



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from bmcmanager.commands.base import (
    BMCManagerServerCommand,
    BMCManagerServerGetCommand
)


def ipmi_address_arguments(parser):
    parser.add_argument(
        '--address-type', choices=['ipv4', 'mac'], required=True,
        help='address type')
    parser.add_argument(
        '--scheme', choices=['http', 'https', ''], default='https',
        help='scheme to prepend to the IPMI address')
    parser.add_argument(
        '--domain', help='domain name to append, only with address type "mac"')


class Get(BMCManagerServerGetCommand):
    """
    print IPMI address
    """
    oob_method = 'get_ipmi_address'

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        ipmi_address_arguments(parser)
        return parser


class Refresh(BMCManagerServerCommand):
    """
    refresh IPMI address on DCIM
    """
    oob_method = 'refresh_ipmi_address'

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        ipmi_address_arguments(parser)
        return parser



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from bmcmanager.commands.base import BMCManagerServerCommand


class PowerOn(BMCManagerServerCommand):
    """
    power on server
    """
    oob_method = 'power_on'


class PowerOff(BMCManagerServerCommand):
    """
    power off server
    """
    oob_method = 'power_off'

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--force', action='store_true', default=False,
            help='force power off the server')
        parser.add_argument(
            '--wait', action='store_true', default=False,
            help='wait for server to power off')
        return parser


class PowerCycle(BMCManagerServerCommand):
    """
    power cycle server
    """
    oob_method = 'power_cycle'


class PowerReset(BMCManagerServerCommand):
    """
    power reset
    """
    oob_method = 'power_reset'


class PowerStatus(BMCManagerServerCommand):
    """
    print server power status
    """
    oob_method = 'power_status'


class LockSwitch(BMCManagerServerCommand):
    """
    lock power switch
    """
    oob_method = 'lock_power_switch'


class UnlockSwitch(BMCManagerServerCommand):
    """
    unlock power switch
    """
    oob_method = 'unlock_power_switch'



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from bmcmanager.commands.base import BMCManagerServerCommand, BMCManagerServerListCommand


class ListSecrets(BMCManagerServerListCommand):
    """
    print available secrets from NetBox
    """
    dcim_fetch_secrets = False
    oob_method = 'get_secrets'


class SetSecret(BMCManagerServerCommand):
    """
    set a secret on NetBox
    """
    dcim_fetch_secrets = False
    oob_method = 'set_secret'

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--secret-role', type=str, required=True,
            help='role for the new NetBox secret')
        parser.add_argument(
            '--secret-name', type=str, required=True,
            help='name for the new NetBox secret')
        parser.add_argument(
            '--secret-plaintext', type=str, required=True,
            help='secret value for the NetBox secret, as plaintext')
        return parser



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import configparser
import os
import sys

from bmcmanager.logs import log


def format_config(config):
    # Recursive function that converts a
    # configparser.ConfigParser object into a dict
    # this basically recursivelly calls dict() on every subdict
    # TODO: refactor
    keys = dict(config).keys()
    formatted = {}
    for k in keys:
        formatted[k.lower()] = config[k]
        if not type(config[k]) == str:
            formatted[k.lower()] = format_config(config[k])
    return formatted


def get_config(config_path):
    try:
        extra_paths = []
        if os.getenv('SNAP_COMMON'):
            extra_paths.extend([os.path.expandvars('$SNAP_COMMON/bmcmanager')])

        if os.getenv('XDG_CONFIG_HOME'):
            extra_paths.extend([
                os.path.expandvars('$XDG_CONFIG_HOME/.config/bmcmanager'),
                os.path.expandvars('$XDG_CONFIG_HOME/bmcmanager')])

        config = configparser.ConfigParser()
        which = config.read([
            config_path,
            os.getenv('BMCMANAGER_CONFIG', ''),
            os.path.expanduser('~/.config/bmcmanager'),
            '/etc/bmcmanager',
            *extra_paths,
        ])

        log.debug('Loaded config from {}'.format(which))

    except configparser.ParsingError as e:
        log.error('Invalid configuration file: {}'.format(e))
        sys.exit(1)

    return format_config(config)



# Original file:
# https://github.com/paramiko/paramiko/blob/master/demos/interactive.py

# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Copyright (C) 2003-2007  Robey Pointer <robeypointer@gmail.com>
#
# This file is part of paramiko.
#
# Paramiko is free software; you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# Paramiko is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Paramiko; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.


import socket
import sys
import select
from paramiko.py3compat import u

import termios
import tty


def posix_shell(chan):
    oldtty = termios.tcgetattr(sys.stdin)
    try:
        tty.setraw(sys.stdin.fileno())
        tty.setcbreak(sys.stdin.fileno())
        chan.settimeout(0.0)

        while True:
            r, w, e = select.select([chan, sys.stdin], [], [])
            if chan in r:
                try:
                    x = u(chan.recv(1024))
                    if len(x) == 0:
                        sys.stdout.write('\r\n*** EOF\r\n')
                        break
                    sys.stdout.write(x)
                    sys.stdout.flush()
                except socket.timeout:
                    pass
            if sys.stdin in r:
                x = sys.stdin.read(1)
                if len(x) == 0:
                    break
                chan.send(x)

    finally:
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, oldtty)



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from cliff.lister import Lister

from bmcmanager.commands.base import BMCManagerServerListCommand, json_argument

LENOVO_RPC_COLUMNS = ['name', 'description', 'response_object']

LENOVO_RPCS = [
    [
        "getactivedircfg",
        "Active Directory auth configuration",
        "WEBVAR_JSONVAR_GETLDAPCFG",
    ],
    [
        "getallcpuinfo",
        "Connected CPUs",
        "WEBVAR_JSONVAR_INVENTORYGETALLCPUINFO",
    ],
    [
        "getalldimminfo",
        "Connected DIMMs",
        "WEBVAR_JSONVAR_INVENTORYGETALLDIMMINFO",
    ],
    [
        "getalllancfg",
        "Get network information for IPMI interfaces",
        "WEBVAR_JSONVAR_GETALLNETWORKCFG",
    ],
    [
        "getallpefcfg",
        "Event Filters",
        "WEBVAR_JSONVAR_HL_GETPEFTABLE",
    ],
    [
        "getallsensors",
        "Get status of all IPMI sensors",
        "WEBVAR_JSONVAR_HL_GETALLSENSORS",
    ],
    [
        "getalluserinfo",
        "IPMI users",
        "WEBVAR_JSONVAR_HL_GETALLUSERINFO",
    ],
    [
        "getanalysedsel",
        "Retrieve SEL, with OEM entries decoded.",
        "WEBVAR_JSONVAR_HL_GETANALYSEDSEL",
    ],
    [
        "getauditlog",
        "Audit log (connections)",
        "WEBVAR_JSONVAR_GETAUDITLOG",
    ],
    [
        "getdatetime",
        "Get Server Time",
        "WEBVAR_JSONVAR_GETDATETIME",
    ],
    [
        "getfruinfo",
        "Fetch all FRU devices",
        "WEBVAR_JSONVAR_HL_GETALLFRUINFO",
    ],
    [
        "gethddinfo",
        "Connected Disks",
        "WEBVAR_JSONVAR_GETINVDRIVEINFO",
    ],
    [
        "gethealthledinfo",
        "Get health LED status",
        "WEBVAR_JSONVAR_GETHEALTHLEDINFO",
    ],
    [
        "getimageinfo",
        "Firmware versions",
        "WEBVAR_JSONVAR_GETIMAGEINFO",
    ],
    [
        "getldapcfg",
        "LDAP auth configuration",
        "WEBVAR_JSONVAR_GETLDAPCFG",
    ],
    [
        "getsysteminfo",
        "BIOS version, Serial number, Model",
        "WEBVAR_JSONVAR_GETSYSTEMINFO",
    ],
    [
        "getuidled",
        "Get identify LED status",
        "WEBVAR_JSONVAR_HL_LED_IDENTIFY_STATE",
    ],
    [
        "hoststatus",
        "Host health",
        "WEBVAR_JSONVAR_HL_SYSTEM_STATE",
    ],
]


class Do(BMCManagerServerListCommand):
    """
    execute an RPC call [Lenovo]
    """
    oob_method = 'lenovo_rpc'

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--rpc', required=True,
            help='name of RPC to execute, see `bmcmanager lenovo rpc list`')
        parser.add_argument('--params', type=json_argument, default={})
        return parser


class List(Lister):
    """
    print known RPC calls [Lenovo]
    """
    def take_action(self, parsed_args):
        return LENOVO_RPC_COLUMNS, LENOVO_RPCS



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import json
import os
import sys

from cliff.command import Command
from cliff.show import ShowOne
from cliff.lister import Lister

from bmcmanager.config import get_config
from bmcmanager.dcim import DCIMS
from bmcmanager.oob import OOBS
from bmcmanager.errors import BMCManagerError
from bmcmanager.logs import log
from bmcmanager import exitcode

README = 'https://github.com/grnet/BMCManager/blob/master/README.md'


def int_in_range_argument(itt):
    """
    argparse integer in range
    """
    def argparse_type(value):
        ivalue = int(value)
        if ivalue not in itt:
            msg = '{} out of valid range [{}..{}]'.format(value, min(itt), max(itt))
            raise argparse.ArgumentTypeError(msg)
        return ivalue

    return argparse_type


def json_argument(arg):
    """
    argparse json argument type
    """
    try:
        return json.loads(arg)
    except json.JSONDecodeError as e:
        raise argparse.ArgumentTypeError('invalid JSON') from e


def base_arguments(parser):
    """
    Base bmcmanager arguments
    """
    parser.add_argument(
        '--config-file',
        help='configuration file path',
        default=''
    )


def server_arguments(parser):
    """
    Add server selection arguments
    """
    parser.add_argument(
        'server',
        help='server name'
    )
    parser.add_argument(
        '-d', '--dcim',
        help='name of DCIM to use',
        choices=['netbox', 'maas'],
        default='netbox',
    )
    parser.add_argument(
        '-t', '--type',
        help='unit type',
        choices=['name', 'rack', 'rack-unit', 'serial'],
        default='search'
    )


def get_dcim(args, config):
    """
    Get a configured DCIM from arguments and configuration
    """
    if args.dcim not in DCIMS:
        raise RuntimeError(
            'Unsupported DCIM "{}", see {}'.format(args.dcim, README))
    if args.dcim not in config:
        raise RuntimeError(
            'No configuration for DCIM "{}", see {}'.format(args.dcim, README))
    return DCIMS[args.dcim](args, config[args.dcim])


def get_oob_config(config, dcim, oob_info, get_secret=True):
    """
    Get configuration for an OOB
    """
    oob_name = oob_info['oob'].lower()
    cfg = {}
    try:
        oob_params = config[oob_name]
    except KeyError:
        raise BMCManagerError('Invalid OOB name {}'.format(oob_name))

    if get_secret and dcim.supports_secrets() and 'credentials' in oob_params:
        secret = dcim.get_secret(oob_params['credentials'], oob_info)
        cfg['username'] = secret['name']
        cfg['password'] = secret['plaintext']
    else:
        cfg['username'] = oob_params['username']
        cfg['password'] = oob_params['password']

    cfg['username'] = os.getenv('BMCMANAGER_USERNAME', cfg['username'])
    cfg['password'] = os.getenv('BMCMANAGER_PASSWORD', cfg['password'])

    cfg['nfs_share'] = os.getenv(
        'BMCMANAGER_NFS_SHARE', oob_params.get('nfs_share'))
    cfg['http_share'] = os.getenv(
        'BMCMANAGER_HTTP_SHARE', oob_params.get('http_share'))

    cfg['oob_params'] = oob_params
    return cfg


def bmcmanager_take_action(cmd, parsed_args):
    cmd.parsed_args = parsed_args
    cmd.config = get_config(parsed_args.config_file)
    dcim = get_dcim(parsed_args, cmd.config)

    idx = None
    for idx, oob_info in enumerate(dcim.get_oobs()):
        oob_config = get_oob_config(cmd.config, dcim, oob_info)
        log.debug('Creating OOB object for {}'.format(oob_info['oob']))
        try:
            oob = OOBS.get(oob_info['oob'])(parsed_args, dcim, oob_config, oob_info)
        except KeyError:
            raise BMCManagerError('Invalid OOB {}'.format(oob_info['oob']))

        try:
            if hasattr(cmd, 'oob_method'):
                return getattr(oob, cmd.oob_method)()
            else:
                return cmd.action(oob)
        except Exception as e:
            log.exception('Unhandled exception: {}'.format(e))

    if idx is None:
        log.fatal('No servers found for "{}"'.format(parsed_args.server))
        return [], []


class BMCManagerServerCommand(Command):
    """
    base command for working with a server
    """
    dcim_fetch_secrets = True

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        base_arguments(parser)
        server_arguments(parser)
        return parser

    def take_action(self, parsed_args):
        bmcmanager_take_action(self, parsed_args)
        sys.exit(exitcode.get())

    def action(self, oob):
        raise NotImplementedError


class BMCManagerServerGetCommand(ShowOne):
    """
    base command for retrieving information for a server
    """
    dcim_fetch_secrets = True

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        base_arguments(parser)
        server_arguments(parser)
        return parser

    def take_action(self, parsed_args):
        return bmcmanager_take_action(self, parsed_args)

    def action(self, oob):
        raise NotImplementedError


class BMCManagerServerListCommand(Lister):
    """
    base command for retrieving a list of information for a server
    """
    dcim_fetch_secrets = True

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        base_arguments(parser)
        server_arguments(parser)
        return parser

    def take_action(self, parsed_args):
        return bmcmanager_take_action(self, parsed_args)

    def action(self, oob):
        raise NotImplementedError



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from bmcmanager import exitcode

OK, WARNING, CRITICAL, UNKNOWN = range(4)
RESULT = {
    OK: 'OK',
    WARNING: 'WARNING',
    CRITICAL: 'CRITICAL',
    UNKNOWN: 'UNKNOWN',
}


def result(status, msg, lines=[], perfdata=[], pre=''):
    if isinstance(msg, list):
        msg = ', '.join(msg)

    print('{} {}: {}'.format(pre, RESULT[status], msg).strip(), end='')
    print(' |' if lines or perfdata else '')

    for line in lines:
        print(line.replace('|', '/'))

    if perfdata:
        print('| ', end='')

    for perf in perfdata:
        print(perf)

    exitcode.update(status)



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from datetime import datetime, timedelta
import re
from subprocess import Popen
import sys
import tempfile
import time
import urllib3

import paramiko
import requests

from bmcmanager.oob.base import OobBase
from bmcmanager.logs import log
from bmcmanager import nagios

from bmcmanager.utils.firmware import version_tuple

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

DISK_TYPE = {
    0: 'HDD',
    1: 'SSD',
}

DISK_INTERFACE = {
    2: 'SAS'
}

DISK_SPEED = {
    3: '6.0Gb/s',
    4: '12.0Gb/s',
}

DISK_STATE = {
    0: 'active',
}

# RPC DEV_TYPE --> osput Device ID
DEV_ID = {
    1: 'TSM',
    2: 'BIOS',
    4: 'PSU',
    5: 'RAID',
    6: 'Mezz',

    201: 'TDM',
    202: 'LG_WIND',
    203: 'LG_LIND',
    205: 'LG_DIAG',

    # Not reported by osput
    204: 'LG_WORK',
    3: 'LG_CPLD_000',
}


class Lenovo(OobBase):
    def _get_console_cookies(self):
        return {'Cookie': 'Language=EN; SessionExpired=true;'}

    def _get_console_headers(self):
        return {'Content-type': 'text/plain'}

    def _get_console_data(self):
        return {
            'WEBVAR_PASSWORD': str(self.password),
            'WEBVAR_USERNAME': str(self.username)
        }

    def _parse_response(self, text):
        """
        Parse response and return (list of results, retryable)
        """
        try:
            start = text.find('[')
            end = text.rfind(']')
            results = eval(text[start:end + 1])

            # drop empty '{}' objects from list of responses
            return [r for r in results if r], False

        except (SyntaxError, TypeError, ValueError):
            if 'session_expired.html' in text:
                log.critical('Failed due to expired session')
                return [], True
            else:
                log.critical('Could not parse response text')
                log.debug('Response was:\n{}'.format(text))
                return [], False

    def _connect(self):
        url = self._get_http_ipmi_host() + self.URL_LOGIN

        cookies = self._get_console_cookies()
        headers = self._get_console_headers()
        data = self._get_console_data()

        self._session = requests.session()

        for count in range(5):
            text = self._post(url, data, cookies, headers).text
            parsed, retryable = self._parse_response(text)
            if not parsed and retryable and count < 4:
                log.info('Will retry after {} seconds'.format(count))
                time.sleep(count)
            else:
                break

        if not parsed:
            log.error('Failed to create session')
            sys.exit(10)

        session_token = parsed[0]['SESSION_COOKIE']
        if session_token == 'Failure_Session_Creation':
            log.error('Probably reached session limit')
            sys.exit(10)

        CSRF_token = parsed[0]['CSRFTOKEN']

        self.session_token = {'SessionCookie': session_token}
        self.CSRF_token = {'CSRFTOKEN': CSRF_token}

    def _post(self, url, data, cookies, headers):
        return self._session.post(
            url,
            data=data,
            cookies=cookies,
            headers=headers,
            verify=False,
            timeout=60
        )

    def console(self):
        self._connect()

        ipmi = self._get_http_ipmi_host()
        url = ipmi + self.URL_VNC.format(ipmi.replace('https://', ''))
        answer = self._post(
            url, None, self.session_token, self.CSRF_token).text

        _, myjviewer = tempfile.mkstemp()
        m = '\n<argument>-title</argument>\n<argument>{}</argument>'
        m = m.format(self.oob_info['identifier'])
        to_repl = '<argument>35</argument>'
        answer = answer.replace(to_repl, to_repl + m)

        with open(myjviewer, 'w') as f:
            f.write(answer)

        cmd = ['/usr/bin/javaws', myjviewer]
        try:
            if self.parsed_args.print_cmd:
                print(' '.join(cmd))
            else:
                Popen(cmd)
        except Exception as e:
            sys.stderr.write('Could not open Java console. {}\n'.format(e))
            sys.exit(10)

    def _system_ram(self):
        port = 22
        hostname = self.oob_info['ipmi'].replace('https://', '')
        username = self.username
        password = self.password

        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(hostname=hostname, port=port,
                       username=username, password=password)
        _, stdout, stderr = client.exec_command('show admin1/hdwr1/memory1')
        output = stdout.read().decode()

        match = re.findall(r'NumberOfBlocks=(\d+)', output, re.MULTILINE)
        try:
            return int(match[0]) // 1024 // 1024
        except:
            return 0

    def system_ram(self):
        return ('ram_gb',), (self._system_ram(),)

    def check_ram(self):
        pre = '{} installed RAM'.format(self.oob_info['identifier'])
        expected = self.parsed_args.expected

        ram = self._system_ram()
        if (expected is None and ram > 0) \
                or (expected is not None and ram == expected):
            nagios.result(nagios.OK, '{}GB'.format(str(ram)), pre=pre)
        elif expected is None:
            nagios.result(nagios.UNKNOWN, 'Failed to read RAM')
        elif ram < expected:
            nagios.result(nagios.CRITICAL, '{}GB, expected {}GB'.format(
                str(ram), str(expected)), pre=pre)
        elif ram > expected:
            nagios.result(nagios.WARNING, '{}GB, expected {}GB'.format(
                str(ram), str(expected)), pre=pre)

    def lock_power_switch(self):
        self._execute(['raw', '0x00', '0x0a', '0x01'])

    def unlock_power_switch(self):
        self._execute(['raw', '0x00', '0x0a', '0x00'])

    def _get_rpc(self, rpc, item='', params=None):
        if not hasattr(self, 'CSRF_token'):
            self._connect()

        ipmi_host = self._get_http_ipmi_host()
        url = ipmi_host + '/rpc/{}.asp'.format(rpc)
        response = self._post(url, params, self.session_token, self.CSRF_token)

        if response.status_code != 200:
            log.critical('Cannot retrieve {}, error {}'.format(
                item, response.status_code))
            return []

        resp, _ = self._parse_response(response.text)
        return resp

    def _get_disks(self):
        return self._get_rpc('gethddinfo', 'disks')

    def _get_sel(self):
        res = self._get_rpc('getanalysedsel', 'SEL', params={
            'WEBVAR_END_RECORD': 65535
        })
        for r in res:
            r['TimeStamp'] = str(
                datetime.utcfromtimestamp(r.get('TimeStamp', 0)))

        return res

    def get_disks(self):
        disks = self._get_disks()
        columns = ['index', 'ctrl', 'slot', 'size_gb', 'type', 'state',
                   'interface', 'speed', 'vendor']
        values = [[
            disk['DRIVE_INDEX'],
            disk['CONTROLLER_INDEX'],
            disk['SLOT_NUMBER'],
            disk['SIZE'] // 1024,
            DISK_TYPE.get(disk['MEDIA_TYPE'], 'N/A'),
            DISK_STATE.get(disk['DEVICE_STATE'], 'unknown'),
            DISK_INTERFACE.get(disk['INTF_TYPE'], 'unknown'),
            DISK_SPEED.get(disk['LINK_SPEED'], 'unknown'),
            disk['VENDOR_ID'],
        ] for disk in disks]

        return columns, values

    def ipmi_logs_analysed(self):
        sel = self._get_sel()
        if not sel:
            return (None, None)

        columns = sel[0].keys()
        values = [[line[col] for col in columns] for line in sel]

        return columns, values

    def _format_disk(self, disk):
        return ' | '.join(map(str, [
            disk['DRIVE_INDEX'],
            'ctrl{}'.format(disk['CONTROLLER_INDEX']),
            'slot{}'.format(disk['SLOT_NUMBER']),
            '{}gb'.format(disk['SIZE'] // 1024),
            DISK_TYPE.get(disk['MEDIA_TYPE'], 'N/A'),
            DISK_STATE.get(disk['DEVICE_STATE'], 'unknown'),
            DISK_INTERFACE.get(disk['INTF_TYPE'], 'unknown'),
            DISK_SPEED.get(disk['LINK_SPEED'], 'unknown'),
            disk['VENDOR_ID'],
        ]))

    def check_disks(self):
        pre = '{} disks'.format(self.oob_info['identifier'])
        expected = self.parsed_args.expected

        disks = self._get_disks()
        disks_ok, disks_crit = [], []
        for d in disks:
            if d['DEVICE_STATE'] != 0 or d['SIZE'] <= 0:
                disks_crit.append(d)
            else:
                disks_ok.append(d)

        status, msg, lines = nagios.OK, [], []
        if disks_crit:
            status = nagios.CRITICAL
            lines.append('{} disks CRITICAL:'.format(len(disks_crit)))
            lines.extend(
                map(lambda d: '- {}'.format(self._format_disk(d)), disks_crit))
            msg.append('{} disks CRITICAL'.format(len(disks_crit)))

        msg.append('{} disks OK'.format(len(disks_ok)))
        if expected is not None and len(disks_ok) != expected:
            msg.append('expected {}'.format(expected))
            status = max(status, nagios.WARNING)

            lines.append('{} disks OK:'.format(len(disks_ok)))
            lines.extend(
                map(lambda d: '- {}'.format(self._format_disk(d)), disks_ok))
            if len(disks_ok) < expected:
                status = max(status, nagios.CRITICAL)

        nagios.result(status, msg, lines, pre=pre)

    def _get_image_info(self):
        return self._get_rpc('getimageinfo', 'firmware versions')

    def _format_fw(self, fw):
        return '- ' + ' | '.join([
            DEV_ID.get(fw['DEV_TYPE'], 'unknown'),
            fw['CURIMG_VER'],
        ])

    def get_firmware(self):
        firmwares = self._get_image_info()
        if not firmwares:
            return (None, None)

        columns = firmwares[0].keys()
        values = [[f[col] for col in columns] for f in firmwares]

        return columns, values

    def refresh_firmware(self):
        custom_fields = {}
        firmwares = self._get_image_info()

        psus = []
        for firmware in firmwares:
            device_id = DEV_ID[firmware['DEV_TYPE']]
            version = firmware['CURIMG_VER']
            if device_id == 'BIOS':
                custom_fields['BIOS'] = version
            elif device_id == 'TSM':
                custom_fields['TSM'] = version
            elif device_id == 'PSU':
                psus.append('{}/{}: {}'.format(
                    firmware['SLOT_NO'], firmware['DEV_IDENTIFIER'], version))

        custom_fields['PSU'] = ', '.join(sorted(psus))

        log.info('Patching custom fields: {}'.format(custom_fields))
        if not self.dcim.set_custom_fields(self.oob_info, custom_fields):
            log.error('Failed to refresh DCIM firmware versions')

    def _matching(self, d1, d2):
        return d1['DEV_IDENTIFIER'] == d2['DEV_IDENTIFIER']

    def firmware_upgrade_osput(self):
        ipmi = self.oob_info['ipmi'].replace('https://', '')
        return self._execute_cmd([
            self.parsed_args.osput,
            '-H', ipmi,
            '-u', self.username,
            '-p', self.password,
            '-f', self.parsed_args.bundle,
            '-c', 'update'])

    def firmware_upgrade_rpc(self):
        args = self.parsed_args
        handle = None

        if 1 in args.stages:
            log.info('Enter FW update mode')
            r = self._get_rpc('getenterfwupdatemode', params={'FWUPMODE': 1})
            log.debug(r)
            if r and 'HANDLE' in r[0]:
                handle = r[0]['HANDLE']
                log.info('Enter FW update mode: OK')
            else:
                log.fatal('Cannot enter FW update mode')
                sys.exit(-1)

        log.info('Update session handle: {}'.format(handle))

        if 2 in args.stages:
            handle = handle or args.handle
            log.info('Rearm firmware update timer')
            r = self._get_rpc('rearmfwupdatetimer', params={'SESSION_ID': handle})
            log.debug(r)

            if r[0]['NEWSESSIONID'] == handle:
                log.info('Rearm firmware update timer: OK')

        if 3 in args.stages:
            handle = handle or args.handle
            if not hasattr(self, 'CSRF_token'):
                self._connect()

            log.info('Uploading firmware bundle')

            ipmi = self.oob_info['ipmi']
            url = ipmi + '/file_upload_firmware.html'
            r = requests.post(
                url,
                verify=False,
                cookies=self.session_token,
                headers=self.CSRF_token,
                files={
                    'bundle?FWUPSessionid={}'.format(handle): args.bundle
                })

            log.debug(r.status_code)
            if r.status_code == 200:
                log.info('Uploading firmware bundle: OK')

        if 4 in args.stages:
            log.info('Get Bundle Upload Status')
            r = self._get_rpc('getbundleupldstatus')
            log.debug(r)

            if r == []:
                log.info('Get Bundle Upload Status: OK')

        if 5 in args.stages:
            log.info('Validate Bundle')
            r = self._get_rpc('validatebundle', params={'BUNDLENAME': 'bundle_bkp.bdl'})
            log.debug(r)
            if r[0]['STATUS'] == 0:
                log.info('Validate Bundle: OK')

        if 6 in args.stages:
            log.info('Replace Bundle')
            r = self._get_rpc('replacebundlebkp')
            log.debug(r)
            if r[0]['STATUS'] == 0:
                log.info('Replace Bundle: OK')

        if 7 in args.stages:
            log.info('Checking for new firmware')
            r = self._get_rpc('getimageinfo')
            log.debug(r)

            def has_update(x):
                new, cur = x['NEWIMG_VER'], x['CURIMG_VER']
                try:
                    return version_tuple(new) > version_tuple(cur)
                except (TypeError, ValueError):
                    return new > cur

            to_update = next(filter(has_update, r), None)
            if to_update is None:
                log.info('No updates available')
                return
            else:
                log.info('Available update: {}'.format(to_update))

        if 8 in args.stages and to_update:
            handle = handle or args.handle
            log.info('Choose component update')
            r = self._get_rpc('setupdatecomp', params={
                'UPDATE_FLAG': to_update['DEV_TYPE'],
                'UPDATE_CNT': 1,
                'FW_DEVICE_TYPE': to_update['DEV_TYPE'],
                'SLOT_NO': to_update['SLOT_NO'],
                'DEV_IDENTIFIER': to_update['DEV_IDENTIFIER'],
                'SESSION_ID': handle,
            })
            log.debug(r)
            if r == []:
                log.info('Choose component update: OK')

        log.info('Firmware upgrade process started')

        if 9 in args.stages:
            begin = datetime.utcnow()
            while datetime.utcnow() < begin + timedelta(minutes=args.timeout):
                try:
                    r = self._get_rpc('getcompupdatestatus')
                    log.debug(r)
                    dev = next((
                        x for x in r if self._matching(x, to_update)), None)
                    log.debug(dev)
                    progress = (dev or {}).get('UPDATE_PERCENTAGE')
                    if progress is None:
                        log.info('Update in progress')
                    else:
                        log.info('Update progress: {}%'.format(progress))
                        if progress == 100:
                            log.info('Update complete!')
                            break
                except (ConnectionResetError, BrokenPipeError):
                    log.info('Update in progress')

                time.sleep(10)

        if 10 in args.stages:
            handle = handle or args.handle
            log.info('Exit FW update mode')
            r = self._get_rpc('getexitfwupdatemode', params={
                'MODE': 0, 'RNDNO': handle
            })
            log.debug(r)
            if r == []:
                log.info('Exit FW update mode: OK')

        log.info('Done!')

    def lenovo_rpc(self):
        args = self.parsed_args
        response = self._get_rpc(args.rpc, 'RPC call', args.params)

        if not response:
            return (None, None)

        columns = list(response[0].keys())
        values = [[r[col] for col in columns] for r in response]

        return columns, values



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from bmcmanager.commands.base import BMCManagerServerCommand, BMCManagerServerGetCommand


class SSH(BMCManagerServerCommand):
    """
    connect to server with SSH
    """
    oob_method = 'ssh'

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--wait', action='store_true', default=False,
            help='wait for server to turn on before starting SSH shell')
        return parser


class FlushJobs(BMCManagerServerCommand):
    """
    flush server pending jobs
    """
    oob_method = 'flush_jobs'


class Identify(BMCManagerServerCommand):
    """
    turn server identifier LED on/off
    """
    oob_method = 'identify'

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--on', nargs='?', type=int, default=15,
            help='turn on identifier LED')
        group.add_argument(
            '--off', action='store_true', default=False,
            help='turn off identifier LED')
        return parser


class Diagnostics(BMCManagerServerCommand):
    """
    print server diagnostics
    """
    oob_method = 'diagnostics'


class Info(BMCManagerServerGetCommand):
    """
    print server info
    """
    oob_method = 'info'


class Upgrade(BMCManagerServerCommand):
    """
    upgrade server
    """
    oob_method = 'upgrade'


class IdracInfo(BMCManagerServerCommand):
    """
    print idrac info
    """
    oob_method = 'idrac_info'



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys

from cliff.app import App
from cliff.commandmanager import CommandManager

from bmcmanager.version import version_string


class BMCManagerApp(App):
    def __init__(self):
        super(BMCManagerApp, self).__init__(
            description='BMCManager',
            version=version_string,
            command_manager=CommandManager('bmcmanager.entrypoints'),
            deferred_help=True
        )


def main(argv=sys.argv[1:]):
    m = BMCManagerApp()
    return m.run(argv)


if __name__ == '__main__':
    sys.exit(main())



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from bmcmanager.commands.base import (
    get_dcim, get_config, BMCManagerServerListCommand)


class List(BMCManagerServerListCommand):
    """
    print list of available servers
    """
    columns = [
        'id',
        'name',
        'site',
        'serial',
        'manufacturer',
        'device_type',
        'status',
        'ipmi',
    ]

    def take_action(self, parsed_args):
        dcim = get_dcim(parsed_args, get_config(parsed_args.config_file))
        values = [[o['info'][col]
                   for col in self.columns] for o in dcim.get_oobs()]

        return self.columns, values



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re
import sys
import time
import paramiko

from subprocess import Popen

from bmcmanager.oob.base import OobBase
from bmcmanager.logs import log


class Dell(OobBase):
    def console(self):
        ipmi_host = self.oob_info['ipmi']
        try:
            Popen(['moob', '-u', '{}'.format(self.username),
                   '-p', '{}'.format(self.password),
                   '-m', ipmi_host.replace('https://', '')])
        except OSError:
            print('Please run "gem install moob"')
            sys.exit(10)

    def _ssh(self, command):
        # performs command using ssh
        # returns decoded output

        nbytes = 4096
        port = 22
        hostname = self.oob_info['ipmi'].replace('https://', '')
        username = self.username
        password = self.password

        client = paramiko.Transport((hostname, port))
        client.connect(username=username, password=password)
        stdout_data = []
        stderr_data = []
        session = client.open_channel(kind='session')
        session.exec_command(command)
        while True:
            if session.recv_ready():
                stdout_data.append(session.recv(nbytes))
            if session.recv_stderr_ready():
                stderr_data.append(session.recv_stderr(nbytes))
            if session.exit_status_ready():
                break
        output = b''.join(stdout_data)
        session.close()
        client.close()

        return output.decode('utf-8')

    def _find_jid(self, output):
        try:
            return re.search(r'JID_.*', output).group(0)
        except AttributeError:
            print('No Job ID found.\nCommand output: ', output)
            sys.exit(10)

    def _confirm_job(self, jid):
        try:
            re.search(r'Job completed successfully', jid).group(0)
        except AttributeError:
            print('Job did not complete successfully.\nCommand output: ', jid)
            sys.exit(10)

    def diagnostics(self):
        jobqueue_view = 'racadm jobqueue view -i {}'
        output = self._ssh('racadm techsupreport collect')
        jid = self._find_jid(output)
        log.info('Sleeping for 3 minutes to collect the TSR report')
        time.sleep(180)
        view_output = self._ssh(jobqueue_view.format(jid))
        self._confirm_job(view_output)
        output = self._ssh('racadm techsupreport export -l {}'.format(
            self.nfs_share))
        jid = self._find_jid(output)
        view_output = self._ssh(jobqueue_view.format(jid))
        self._confirm_job(view_output)

    def autoupdate(self):
        enable_updates_output = self._ssh(
            'racadm set lifecycleController.lcattributes.AutoUpdate Enabled')
        schedule_updates_output = self._ssh(
            'racadm autoupdatescheduler create -l {} '
            '-f grnet_1.00_Catalog.xml -a 0 -time 08:30 '
            '-dom * -wom * -dow * -rp 1'.format(self.http_share))
        print(enable_updates_output)
        print(schedule_updates_output)

    def upgrade(self):
        http_addr = self.http_share.strip('http:/')
        self._ssh('racadm update -f {} -e {} -t HTTP -a FALSE'.format(
            'grnet_1.00_Catalog.xml', http_addr))

    def idrac_info(self):
        firm_info = 'racadm get idrac.info'
        bios_info = 'racadm get bios.sysinformation'
        print(self._ssh(firm_info))
        print(self._ssh(bios_info))

    def clear_autoupdate(self):
        clear_command = 'racadm autoupdatescheduler clear'
        print(self._ssh(clear_command))

    def flush_jobs(self):
        flush_command = 'racadm jobqueue delete --all'
        print(self._ssh(flush_command))

    def pdisks_status(self):
        pdisks_status_command = 'racadm storage get pdisks -o'
        print(self._ssh(pdisks_status_command))

    def storage_status(self):
        storage_status_command = 'racadm storage get status'
        print(self._ssh(storage_status_command))

    def controllers_status(self):
        controllers_status_command = 'racadm storage get controllers -o'
        print(self._ssh(controllers_status_command))



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from bmcmanager.commands.base import BMCManagerServerCommand, BMCManagerServerListCommand


class Get(BMCManagerServerListCommand):
    """
    print attached disks
    """
    oob_method = 'get_disks'


class Check(BMCManagerServerCommand):
    """
    check attached disks [Nagios]
    """
    oob_method = 'check_disks'

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--expected', type=int, default=None,
            help='number of healthy disks the server must have')
        return parser



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re
from subprocess import Popen
from urllib import error as urlerror
from urllib import request
import tempfile

from bs4 import BeautifulSoup

from bmcmanager.oob.base import OobBase


class Fujitsu(OobBase):
    def _get_realm(self):
        """Do an unauthenticated request to get the realm for auth."""
        opener = request.build_opener()
        request.install_opener(opener)
        try:
            request.urlopen(self.oob_info['ipmi'])
        except urlerror.HTTPError as err:
            header = err.headers.get('WWW-Authenticate')
            m = re.match(r'Digest realm="([@\w\s-]+)",', header)
            realm = m.groups()[0]
        return realm

    def _install_auth(self):
        """Setup digest auth"""
        realm = self._get_realm()

        uri = self.oob_info['ipmi']
        username = self.username
        password = self.password

        auth_handler = request.HTTPDigestAuthHandler()
        auth_handler.add_password(realm=realm,
                                  uri=uri, user=username, passwd=password)
        opener = request.build_opener(auth_handler)
        request.install_opener(opener)

    def _find_avr_url(self):
        """Parse the main page to find the URL for JWS"""
        url = self.oob_info['ipmi']

        req = request.Request(url)
        data = request.urlopen(req).read()
        soup = BeautifulSoup(data)
        jnlp_desc = [u'Video Redirection (JWS)']
        links = soup.find_all('a', href=True)
        for link in links:
            if link.contents == jnlp_desc:
                return url + '/' + link['href']

    def _save_tmp_jnlp(self):
        """Fetch the XML jnlp file and save in tmp"""
        avr_url = self._find_avr_url()
        resource = request.urlopen(request.Request(avr_url))
        xml_data = resource.read().decode('utf-8')
        _, tmppath = tempfile.mkstemp()
        with open(tmppath, 'w') as tmpfile:
            tmpfile.write(xml_data)
        return tmppath

    def console(self):
        self._install_auth()
        avr_file = self._save_tmp_jnlp()
        Popen(['/usr/bin/javaws', avr_file])



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from bmcmanager.commands.base import (
    BMCManagerServerCommand,
    BMCManagerServerGetCommand
)


class Get(BMCManagerServerGetCommand):
    """
    print IPMI credentials
    """
    oob_method = 'creds'


class Set(BMCManagerServerCommand):
    """
    change user IPMI password
    """
    oob_method = 'set_ipmi_password'

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--secret-role', type=str, required=False,
            help='set the password as a NetBox secret with this role')
        parser.add_argument(
            '--new-password', type=str, required=True,
            help='new IPMI password')
        return parser



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import sys
import urllib.request
from subprocess import call, CalledProcessError

from cliff.lister import Lister
from cliff.command import Command

from bmcmanager import nagios, exitcode
from bmcmanager.commands.base import (
    BMCManagerServerCommand,
    BMCManagerServerListCommand,
    int_in_range_argument
)
from bmcmanager.logs import log
from bmcmanager.firmwares import firmware_fetchers


class Get(BMCManagerServerListCommand):
    """
    print server firmware versions
    """
    oob_method = 'get_firmware'


class Refresh(BMCManagerServerCommand):
    """
    update server firmware information on DCIM
    """
    oob_method = 'refresh_firmware'


class Check(BMCManagerServerCommand):
    """
    check server firmware version [Nagios]
    """
    oob_method = 'check_firmware'
    dcim_fetch_secrets = False


class UpgradeRPC(BMCManagerServerCommand):
    """
    perform firmware upgrade using RPC
    """
    oob_method = 'firmware_upgrade_rpc'
    all_stages = range(1, 11)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--timeout', type=int, default=60,
            help='advanced; seconds before failing because of timeout')
        parser.add_argument(
            '--handle', type=int, default=None,
            help='advanced; Use this handle for upgrade [Lenovo]')
        parser.add_argument(
            '--bundle', required=True, type=argparse.FileType('rb'),
            help='Bundle file to use for firmware upgrade')
        parser.add_argument(
            '--stages', nargs='+', default=self.all_stages,
            type=int_in_range_argument(self.all_stages),
            help='advanced; Only perform specific upgrade stages [Lenovo]')

        return parser


class UpgradeOsput(BMCManagerServerCommand):
    """
    perform firmware upgrade using osput [Lenovo]
    """
    oob_method = 'firmware_upgrade_osput'

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--osput', type=str, default='osput',
            help='override path to the `osput` executable [lenovo]')
        parser.add_argument(
            '--bundle', type=str, required=True,
            help='bundle file to use for firmware upgrade')
        return parser


class LatestGet(Lister):
    """
    print and download latest firmware version bundles
    """
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'model', choices=firmware_fetchers.keys(),
            help='server model for which to look for new firwmare')
        parser.add_argument(
            '--download-to',
            help='download available firmware files to this location')
        parser.add_argument(
            '--innoextract', action='store_true',
            help='extract `.exe` files using innoextract')
        return parser

    def _execute_cmd(self, command):
        log.debug('Executing {}'.format(' '.join(command)))
        try:
            call(command)
        except CalledProcessError as e:
            raise RuntimeError('Command {} failed: {}'.format(
                ' '.join(command), str(e)))

    def take_action(self, parsed_args):
        try:
            fetcher = firmware_fetchers[parsed_args.model]

        except KeyError as e:
            log.error('Unsupported device type: {}'.format(e))
            sys.exit(-1)

        result, downloads = fetcher().get()

        columns = ['component', 'name', 'version', 'date']
        values = [[item[col] for col in columns] for item in result]

        if parsed_args.download_to is None:
            return columns, values

        try:
            os.makedirs(parsed_args.download_to, exist_ok=True)
        except OSError as e:
            log.error('Could not create download directory: {}'.format(e))
            sys.exit(-1)

        for url in downloads:
            name = url[url.rfind('/') + 1:]
            file_name = os.path.join(parsed_args.download_to, name)
            log.info('Downloading {} to {}'.format(url, file_name))
            try:
                with open(file_name, 'wb') as fout:
                    fout.write(urllib.request.urlopen(url).read())
            except (urllib.error.URLError, OSError) as e:
                log.error('Failed: {}'.format(e))

            if parsed_args.innoextract and name.endswith('.exe'):
                log.info('Extracting with innoextract')
                self._execute_cmd([
                    'innoextract', file_name, '-d', parsed_args.download_to])

        return columns, values


class LatestCheck(Command):
    """
    check for new firmware versions
    """
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'model', choices=firmware_fetchers.keys(),
            help='server model for which to look for new firwmare')
        parser.add_argument(
            '--after', type=str, required=True,
            help='check that there are no firmware releases after this date (YYYY-MM-DD)'
        )
        return parser

    def take_action(self, parsed_args):
        try:
            fetcher = firmware_fetchers[parsed_args.model]

        except KeyError as e:
            log.error('Unsupported device type: {}'.format(e))
            sys.exit(-1)

        result, _ = fetcher().get()

        new_firmware = []
        for item in result:
            if item['date'] > parsed_args.after:
                new_firmware.append('- {} / {} / {} / {}'.format(
                    item['date'],
                    item['component'],
                    item['version'],
                    item['name'],
                ))

        if new_firmware:
            state = nagios.CRITICAL
            msg = '{} new firmware version releases since {}'.format(
                len(new_firmware), parsed_args.after)
        else:
            state = nagios.OK
            msg = 'No new firmware version releases since {}'.format(
                parsed_args.after)

        nagios.result(
            state, msg, lines=new_firmware, pre='Firmware Versions')

        sys.exit(exitcode.get())



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import sys
import os

import requests

from bmcmanager.logs import log
from bmcmanager.dcim.base import DcimBase, DcimError


class Netbox(DcimBase):
    def __init__(self, args, config):
        super(Netbox, self).__init__(args, config)

        self.timeout = 10
        try:
            timeout = self.dcim_params.get('timeout', self.timeout)
            self.timeout = int(timeout)
        except (TypeError, ValueError):
            log.warning('Ignoring invalid timeout: {}'.format(timeout))

        self.device_type_ids = None
        raw_ids = self.dcim_params.get('device_type_id')
        if raw_ids is not None:
            try:
                # '1, 3,2' --> [1, 2, 3]
                self.device_type_ids = list(
                    map(int, map(str.strip, raw_ids.split(','))))
            except (TypeError, ValueError):
                log.warning('Ignoring invalid device type ids: {}'.format(
                    raw_ids))

        self.info = self._retrieve_info()

    def _get_params(self):
        if self.is_serial:
            return {'serial': self.identifier}
        elif self.is_rack_unit:
            return {'name': self.identifier}
        elif self.is_rack:
            return {'rack_id': self._get_rack_id()}

        params = {'q': self.identifier}
        if self.device_type_ids is not None:
            params['device_type_id'] = self.device_type_ids
        return params

    def _get_rack_id(self):
        log.debug('Querying the Netbox API for rack {}'.format(
            self.identifier))
        url = os.path.join(self.api_url, 'api/dcim/racks/')
        params = {'name': self.identifier}
        json_response = self._do_request(url, params)

        log.debug('Decoding the response')
        # we expect the response to be a json object
        response = json_response.json()
        if len(response['results']) != 1:
            raise DcimError('Did not find valid results for rack {}'.format(
                self.identifier))
        return response['results'][0]['id']

    def _get_headers(self, with_session_key=False):
        headers = {'Accept': 'application/json'}

        token = self.dcim_params.get('netbox_token')
        if token is not None:
            headers.update({'Authorization': 'Token {}'.format(token)})

        if with_session_key:
            session_key = self.dcim_params.get('session_key')
            if session_key is not None:
                headers.update({'X-Session-Key': session_key})

        return headers

    def _do_request(self, url, params, with_session_key=False, method='get'):
        headers = self._get_headers(with_session_key)

        log.debug('HTTP {} {}, {}, {}'.format(
            method.upper(), url, str(params), str(headers)))
        try:
            f = getattr(requests, method)
            return f(url, params=params, headers=headers, timeout=self.timeout)
        except (
            requests.exceptions.Timeout,
            requests.exceptions.ConnectionError):
            # useful instead of a long exception dump
            sys.stderr.write('Request timed out {}'.format(url))
            exit(1)

    def _retrieve_info(self):
        log.debug('Querying the Netbox API for {}'.format(self.identifier))
        url = os.path.join(self.api_url, 'api/dcim/devices/')
        params = self._get_params()
        params['limit'] = 0
        json_response = self._do_request(url, params)
        log.debug('Decoding the response')
        # we expect the response to be a json object
        return json_response.json()

    def get_short_info(self, result):
        return {
            'id': result['id'],
            'name': result['name'],
            'display_name': result['display_name'],
            'serial': result['serial'],
            'ipmi': result['custom_fields']['IPMI'] or 'unknown-address',
            'manufacturer': result['device_type']['manufacturer']['slug'],
            'device_type': result['device_type']['slug'],
            'bios': result['custom_fields']['BIOS'],
            'tsm': result['custom_fields']['TSM'],
            'psu': result['custom_fields']['PSU'],
            'site': result['site']['name'],
            'status': result['status']['label'],
        }

    def get_info(self):
        return self.info

    def get_oobs(self):
        for result in self.info['results']:
            yield {
                'asset_tag': result['asset_tag'],
                'ipmi': result['custom_fields']['IPMI'],
                'oob': result['device_type']['manufacturer']['slug'],
                'info': self.get_short_info(result),
                'identifier': result['name'],
                'custom_fields': result['custom_fields'],
            }

    def get_secret(self, role, oob_info):
        device = oob_info['info']['name']
        log.debug('Querying secret {} of device {}'.format(role, device))
        response = self._do_request(
            url=os.path.join(self.api_url, 'api/secrets/secrets/'),
            params={'role': role, 'device': device.upper()},
            with_session_key=True)

        try:
            return response.json()['results'][0]

        except (TypeError, KeyError, IndexError):
            log.warning('Did not find secret {} for device {}'.format(
                role, device))
            return {
                'name': None,
                'plaintext': None,
            }

    def _get_secret_role_id(self, role_name):
        log.debug('Searching for id of secret role {}'.format(role_name))
        response = self._do_request(
            url=os.path.join(self.api_url, 'api/secrets/secret-roles/'),
            params={'slug': role_name}
        )

        try:
            return response.json()['results'][0]['id']
        except (TypeError, KeyError, IndexError):
            return None

    def set_secret(self, role_name, oob_info, secret_name, secret_text):
        log.debug('Will upsert secret {} for device {}'.format(
            role_name, oob_info['name']))

        role_id = self._get_secret_role_id(role_name)
        if role_id is None:
            log.critical('unknown role slug {}'.format(role_name))

        url = os.path.join(self.api_url, 'api/secrets/secrets/')
        f = requests.post

        existing_secrets = self.get_secrets(oob_info)
        for s in existing_secrets:
            if s['role'] == role_name and s['name'] == secret_name:
                log.debug('Updating secret with role {} and name {}'.format(
                    role_name, secret_name))
                url = os.path.join(url, '{}/'.format(s['id']))
                f = requests.patch
                break

        return f(
            url=url,
            headers=self._get_headers(with_session_key=True),
            json={
                'device': oob_info['id'],
                'role': role_id,
                'name': secret_name,
                'plaintext': secret_text,
            },
        )

    def _get_secrets(self, oob_info):
        log.debug('Searching for secrets of device {}'.format(
            oob_info['name']))

        response = self._do_request(
            url=os.path.join(self.api_url, 'api/secrets/secrets/'),
            params={
                'device': oob_info['name'],
            },
            with_session_key=True
        )

        return response

    def get_secrets(self, oob_info):
        resp = self._get_secrets(oob_info)
        try:
            return [{
                'id': s['id'],
                'name': s['name'],
                'role': s['role']['slug'],
                'plaintext': s['plaintext'],
            } for s in resp.json()['results']]
        except (KeyError, ValueError, TypeError):
            return []

    def set_custom_fields(self, oob_info, custom_fields):
        return requests.patch(
            url=os.path.join(self.api_url, 'api/dcim/devices/{}/'.format(oob_info['info']['id'])),
            headers=self._get_headers(),
            json={
                'custom_fields': custom_fields,
            },
        ).status_code == 200

    def oob_url(self, oob_info):
        return os.path.join(self.api_url, 'dcim/devices/{}/'.format(oob_info['info']['id']))



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os
import re
from subprocess import Popen, check_output, CalledProcessError, call
import sys

import paramiko

from bmcmanager.interactive import posix_shell
from bmcmanager.utils import firmware
from bmcmanager import nagios
from bmcmanager.logs import log

if sys.platform == 'darwin':
    BROWSER_OPEN = 'open'
else:
    BROWSER_OPEN = 'xdg-open'

# Column indexes for ipmi-sel output
SEL_ID, SEL_DATE, SEL_TIME, SEL_NAME, SEL_TYPE, SEL_STATE, SEL_EVENT = range(7)

# Column indexes for ipmi-sensors output
SSR_ID, SSR_NAME, SSR_TYPE, SSR_STATE, SSR_VALUE, SSR_UNIT, _, SSR_CRITL, \
    SSR_WARNL, SSR_WARNH, SSR_CRITH, _, SSR_DESC = range(13)


class OobBase(object):
    """
    Base OOB class
    """
    URL_LOGIN = '/rpc/WEBSES/create.asp'
    URL_VNC = '/Java/jviewer.jnlp?EXTRNIP={}&JNLPSTR=JViewer'

    def __init__(self, parsed_args, dcim, oob_config, oob_info):
        self.parsed_args = parsed_args
        self.oob_info = oob_info
        self.dcim = dcim
        self.oob_config = oob_config
        self.username = self.oob_config['username']
        self.password = self.oob_config['password']
        self.nfs_share = self.oob_config['nfs_share']
        self.http_share = self.oob_config['http_share']

    def _print(self, msg):
        sys.stdout.write('{}:\n{}\n'.format(self.oob_info['identifier'], msg))

    def info(self):
        log.debug('Executing info')
        info = self.oob_info['info']

        columns = info.keys()
        values = [info[col] for col in columns]

        return columns, values

    def _execute_popen(self, command):
        log.debug('Executing {}'.format(' '.join(command)))
        try:
            Popen(command)
        except:
            raise OobError('Could not open browser {}'.format(
                ' '.join(command)))

    def _get_http_ipmi_host(self):
        ipmi_host = self.oob_info['ipmi']
        if not ipmi_host.startswith('http'):
            ipmi_host = 'https://{}'.format(ipmi_host)

        return ipmi_host

    def open(self):
        self._execute_popen([BROWSER_OPEN, self._get_http_ipmi_host()])

    def ssh(self):
        status_command = ['chassis', 'power', 'status']
        if self.parsed_args.wait:
            if 'off' in self._execute(status_command, output=True):
                log.info('Waiting for machine to turn on...')

            while (1):
                if 'off' not in self._execute(status_command, output=True):
                    break

        host = self.oob_info['asset_tag']
        if not host:
            raise OobError('Cannot perform ssh without an asset tag')
        call(['ssh', host])

    def _get_ipmi_tool_prefix(self):
        host = self.oob_info['ipmi'].replace('https://', '')
        return ['ipmitool', '-U', self.username, '-P', self.password,
                '-I', 'lanplus', '-H', host]

    # command is an array
    def _execute(self, command, output=False):
        if not self.oob_info['ipmi']:
            log.warn('No IPMI field for {}'.format(self.oob_info['oob']))
            return ''

        prefix = self._get_ipmi_tool_prefix()
        command = prefix + command

        return self._execute_cmd(command, output)

    # command is an array
    def _execute_cmd(self, command, output=False):
        log.debug('Executing {}'.format(' '.join(command)))
        try:
            if output:
                return check_output(command).decode('utf-8')

            call(command)
        except CalledProcessError as e:
            raise OobError('Command {} failed: {}'.format(
                ' '.join(command), str(e)))
        except UnicodeError as e:
            raise OobError('Decoding output of {} failed: {}'.format(
                ' '.join(command), str(e)))

    def identify(self):
        if self.parsed_args.off:
            arg = 0
        else:
            arg = self.parsed_args.on or 'force'

        self._print(self._execute(
            ['chassis', 'identify', str(arg)], output=True).strip())

    def status(self):
        lines = self._execute(
            ['chassis', 'status'], output=True).strip().split('\n')
        columns, values = [], []
        for line in lines:
            try:
                key, value = line.split(':')
            except ValueError:
                continue

            columns.append(key.strip())
            values.append(value.strip())

        return columns, values

    def power_status(self):
        self._print(self._execute(
            ['chassis', 'power', 'status'],
            output=True
        ).strip())

    def power_on(self):
        self._execute(['chassis', 'power', 'on'])

    def power_off(self):
        cmd = ['chassis', 'power']
        if self.parsed_args.force:
            cmd.append('off')
        else:
            cmd.append('soft')
        self._execute(cmd)
        if self.parsed_args.wait:
            while 1:
                stdout = self._execute(
                    ['chassis', 'power', 'status'], output=True)
                if 'off' in stdout:
                    break

    def power_cycle(self):
        self._execute(['chassis', 'power', 'cycle'])

    def power_reset(self):
        self._execute(['chassis', 'power', 'reset'])

    def boot_pxe(self):
        self._execute(['chassis', 'bootdev', 'pxe'])

    def boot_local(self):
        self._execute(['chassis', 'bootdev', 'disk'])

    def ipmi_reset(self):
        cmd = ['mc', 'reset']
        if self.parsed_args.force:
            cmd.append('cold')
        else:
            cmd.append('warm')

        self._execute(cmd)

    def _ipmi_logs_cmd(self, args=[]):
        return ['sel', 'list', *args]

    def ipmi_logs(self):
        lines = self._execute(
            self._ipmi_logs_cmd(), output=True).strip().split('\n')

        columns = ['id', 'date', 'time', 'name', 'event', 'state']
        values = [list(map(str.strip, line.split('|'))) for line in lines]

        return columns, values

    def clear_ipmi_logs(self):
        self._print(self._execute(['sel', 'clear'], output=True).strip())

    def console(self):
        raise NotImplementedError('console')

    def diagnostics(self):
        raise NotImplementedError('diagnostics')

    def autoupdate(self):
        raise NotImplementedError('autoupdate')

    def upgrade(self):
        raise NotImplementedError('upgrade')

    def idrac_info(self):
        raise NotImplementedError('idrac-info')

    def remove_autoupdate(self):
        raise NotImplementedError('remove-autoupdate')

    def flush_jobs(self):
        raise NotImplementedError('flush-jobs')

    def pdisks_status(self):
        raise NotImplementedError('pdisks-status')

    def storage_status(self):
        raise NotImplementedError('storage-status')

    def controllers_status(self):
        raise NotImplementedError('controllers-status')

    def system_ram(self):
        raise NotImplementedError('system-ram')

    def lock_power_switch(self):
        raise NotImplementedError('lock-power-switch')

    def unlock_power_switch(self):
        raise NotImplementedError('unlock-power-switch')

    def _ipmi_sensors_cmd(self, host, username, password, args=[]):
        if os.getenv('XDG_CACHE_HOME'):
            cache_dir_arg = '--sdr-cache-dir=$XDG_CACHE_HOME'
            args = [*args, os.path.expandvars(cache_dir_arg)]

        return [
            'ipmi-sensors', '-h', host,
            '-u', username, '-p', password, '-l', 'user',
            '--quiet-cache', '--sdr-cache-recreate',
            '--interpret-oem-data', '--output-sensor-state',
            '--ignore-not-available-sensors',
            '--driver-type=LAN_2_0',
            '--output-sensor-thresholds',
            *args]

    def _ipmi_sel_cmd(self, host, username, password, args=[]):
        return [
            'ipmi-sel', '-h', host,
            '-u', username, '-p', password, '-l', 'user',
            '--driver-type=LAN_2_0',
            '--output-event-state',
            '--interpret-oem-data',
            '--entity-sensor-names',
            '--sensor-types=all',
            '--ignore-sdr-cache',
            *args
        ]

    def _ipmi_dcmi_cmd(self, host, username, password, args=[]):
        return [
            'ipmi-dcmi', '-h', host,
            '-u', username, '-p', password, '-l', 'user',
            '--driver-type=LAN_2_0',
            '--get-system-power-statistics',
            *args
        ]

    def ipmi_sensors(self):
        host = self.oob_info['ipmi'].replace('https://', '')
        cmd = self._ipmi_sensors_cmd(host, self.username, self.password)
        lines = self._execute_cmd(cmd, output=True).strip().split('\n')
        columns = list(map(str.strip, lines[0].split('|')))
        values = [list(map(str.strip, line.split('|'))) for line in lines[1:]]

        return (columns, values)

    def _format_sensor(self, sensor):
        return '- ' + ' | '.join([sensor[x] for x in [
            SSR_ID, SSR_TYPE, SSR_NAME, SSR_VALUE, SSR_UNIT, SSR_DESC
        ]])

    def __clear_na(self, x):
        return '' if x == 'N/A' else x

    def _format_sensor_perfdata(self, sensor):
        if sensor[SSR_VALUE] == 'N/A':
            return ''

        warning = ''
        warning_low = self.__clear_na(sensor[SSR_WARNL])
        warning_high = self.__clear_na(sensor[SSR_WARNH])
        if warning_low or warning_high:
            warning = '{}:{}'.format(warning_low, warning_high)

        critical = ''
        critical_low = self.__clear_na(sensor[SSR_CRITL])
        critical_high = self.__clear_na(sensor[SSR_CRITH])
        if critical_low or critical_high:
            critical = '{}:{}'.format(critical_low, critical_high)

        result = "'{}'={}".format(sensor[SSR_NAME], sensor[SSR_VALUE])
        if warning or critical:
            result = '{};{};{}'.format(result, warning, critical)

        return result

    def _format_sel(self, sel):
        return '- ' + ' | '.join([sel[x] for x in [
            SEL_ID, SEL_DATE, SEL_TIME, SEL_TYPE, SEL_NAME, SEL_EVENT
        ]])

    def _get_sel_errors(self, host):
        cmd = self._ipmi_sel_cmd(host, self.username, self.password)
        logs = self._execute_cmd(cmd, output=True)
        for line in reversed(logs.split('\n')[1:-1]):
            split = list(map(lambda x: x.strip(), line.split('|')))
            if split[SEL_STATE] != 'Nominal':
                yield split

    def _sel_is_firmware_upgrade(self, line):
        checks = {
            SEL_DATE: 'PostInit',
            SEL_TIME: 'PostInit',
            SEL_TYPE: 'Version Change',
        }
        return all(line[k] == v for k, v in checks.items())

    def check_ipmi(self):
        pre = '{} IPMI Status'.format(self.oob_info['identifier'])
        try:
            host = self.oob_info['ipmi'].replace('https://', '')
        except:
            nagios.result(nagios.UNKNOWN, 'No IPMI information', pre=pre)
            return

        perfdata = []
        cmd = self._ipmi_dcmi_cmd(host, self.username, self.password)
        try:
            dcmi_output = self._execute_cmd(cmd, output=True)
        except OobError:
            nagios.result(nagios.UNKNOWN, 'ipmi-dcmi failed', pre=pre)
            return

        match = re.findall(r'Current Power\s*:\s*(\d+)', dcmi_output)
        if match:
            perfdata.append("'Current Power'={}".format(match[0]))

        sel_errors = list(self._get_sel_errors(host))

        cmd = self._ipmi_sensors_cmd(host, self.username, self.password)
        try:
            sensors = self._execute_cmd(cmd, output=True)
        except OobError:
            nagios.result(nagios.UNKNOWN, 'ipmi-sensors failed', pre=pre)
            return

        sensor_warnings = []
        sensor_errors = []
        for line in sensors.split('\n')[1:-1]:
            split = list(map(lambda x: x.strip(), line.split('|')))

            data = self._format_sensor_perfdata(split)
            if data:
                perfdata.append(data)

            if split[SSR_STATE] in ['Nominal', 'N/A']:
                continue
            elif split[SSR_STATE] == 'Warning':
                sensor_warnings.append(split)
            else:
                sensor_errors.append(split)

        status, msg, lines = nagios.OK, [], []
        if sensor_warnings:
            status = nagios.WARNING
            msg.append('{} sensors warning'.format(len(sensor_warnings)))
            lines.extend([
                'Warning sensors:',
                *map(lambda x: self._format_sensor(x), sensor_warnings)])
        if sensor_errors:
            status = nagios.CRITICAL
            msg.append('{} sensors critical'.format(len(sensor_errors)))
            lines.extend([
                'Critical sensors:',
                *map(lambda x: self._format_sensor(x), sensor_errors)])
        if sel_errors:
            status = nagios.CRITICAL
            msg.append('{} SEL entries'.format(len(sel_errors)))
            sel_entries_header = 'SEL entries:'

            # cap number of SEL entries returned
            if len(sel_errors) > 10:
                sel_entries_header += ' (showing latest 10/{})'.format(
                    len(sel_errors))
                sel_errors = sel_errors[:10]

            lines.extend([
                sel_entries_header,
                *map(lambda x: self._format_sel(x), sel_errors)])

        perfdata = [' '.join(perfdata)]
        nagios.result(status, msg or 'SEL, Sensors OK', lines, perfdata, pre)

    def creds(self):
        ipmi = (self.oob_info['ipmi'] or '').replace('https://', '')
        columns = ('address', 'username', 'password')
        values = (ipmi, self.username, self.password)
        return columns, values

    def ipmi_ssh(self):
        port = 22
        hostname = self.oob_info['ipmi'].replace('https://', '')
        username = self.username
        password = self.password

        if self.parsed_args.command:
            client = paramiko.SSHClient()
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            client.connect(hostname=hostname, port=port,
                           username=username, password=password)
            _, out, err = client.exec_command(
                ' '.join(self.parsed_args.command))
            print(out.read().decode())
            if err:
                print(err.read().decode(), file=sys.stderr)
        else:
            client = paramiko.Transport((hostname, port))
            client.connect(username=username, password=password)
            session = client.open_channel(kind='session')
            session.get_pty()
            session.invoke_shell()
            posix_shell(session)

        client.close()

    def check_firmware(self):
        pre = '{} Firmware versions'.format(self.oob_info['identifier'])
        state, msg = firmware.check_firmware(
            self.oob_info['custom_fields'], self.oob_config)

        nagios.result(state, msg, pre=pre)

    def check_ram(self):
        raise NotImplementedError('check-ram')

    def check_disks(self):
        raise NotImplementedError('check-disks')

    def get_disks(self):
        raise NotImplementedError('get-disks')

    def refresh_firmware(self):
        raise NotImplementedError('refresh-firmware')

    def ipmi_logs_analysed(self):
        raise NotImplementedError('ipmi-logs-analysed')

    def get_secrets(self):
        if not self.dcim.supports_secrets():
            log.fatal('Secrets not supported by DCIM')

        secrets = self.dcim.get_secrets(self.oob_info['info'])

        columns = ['id', 'role', 'name', 'plaintext']
        values = [[secret[col] for col in columns] for secret in secrets]
        return columns, values

    def set_secret(self):
        if not self.dcim.supports_secrets():
            log.fatal('Secrets not supported by DCIM')

        r = self.dcim.set_secret(
            self.parsed_args.secret_role,
            self.oob_info['info'],
            self.parsed_args.secret_name,
            self.parsed_args.secret_plaintext)

        if r.status_code >= 300:
            self._print('Error {}'.format(r.text))

    def ipmitool(self):
        self._execute(self.parsed_args.args)

    def set_ipmi_password(self):
        args = self.parsed_args

        stdout = self._execute(['user', 'list'], output=True).split('\n')
        header = stdout[0]
        name_start = header.find('Name')
        found = False
        for line in stdout[1:]:
            uid = line[:name_start].strip()
            username = line[name_start:line.find(' ', name_start)]

            if username == self.username:
                found = True
                log.debug('ID of user {} is {}'.format(username, uid))
                break

        if not found:
            self._print('User {} not found'.format(self.username))
            return

        self._print(self._execute([
            'user', 'set', 'password', uid, args.new_password], output=True))

        if args.secret_role is not None:
            self._print('Updating {} NetBox secret'.format(args.secret_role))
            r = self.dcim.set_secret(args.secret_role, self.oob_info['info'],
                                     username, args.new_password)
            if r.status_code >= 300:
                log.critical('Setting NetBox secret failed')
                log.critical('Error {}:\n{}'.format(r.status_code, r.text))
            else:
                self._print('Successfully updated secret')

    def get_firmware(self):
        raise NotImplementedError('get-firmware')

    def firmware_upgrade(self):
        raise NotImplementedError('firmware-upgrade')

    def lenovo_rpc(self):
        raise NotImplementedError('lenovo-rpc')

    def clear_firmware_upgrade_logs(self):
        host = self.oob_info['ipmi'].replace('https://', '')
        sel_errors = self._get_sel_errors(host)
        if all(self._sel_is_firmware_upgrade(err) for err in sel_errors):
            self.clear_ipmi_logs()

    def _get_ipmi_address(self):
        args = self.parsed_args
        regex = {
            'ipv4': r'^IP Address\s*:\s*(?P<addr>\d+(\.\d+){3})',
            'mac':
                r'MAC Address\s*:\s*(?P<addr>[a-f0-9]{2}(\:[a-f0-9]{2}){5})',
        }.get(args.address_type)

        if regex is None:
            log.error('Unknown address type: {}'.format(args.address_type))
            return ''

        stdout = self._execute(['lan', 'print'], output=True)

        m = next(re.finditer(regex, stdout, re.MULTILINE), None)
        if m is None:
            log.error('No IPMI address found')
            return

        addr = m.group('addr')
        if args.address_type == 'mac':
            addr = addr.replace(':', '').upper()
            if args.domain:
                addr = '{}.{}'.format(addr, args.domain.lstrip('.'))
        if args.scheme:
            addr = '{}://{}'.format(args.scheme, addr)

        return addr

    def refresh_ipmi_address(self):
        addr = self._get_ipmi_address()
        custom_fields = {'IPMI': addr}
        log.info('Patching custom fields: {}'.format(custom_fields))
        if not self.dcim.set_custom_fields(self.oob_info, custom_fields):
            log.error('Failed to refresh IPMI')

    def get_ipmi_address(self):
        addr = self._get_ipmi_address()
        return ('address',), (addr,)

    def open_dcim(self):
        self._execute_popen([BROWSER_OPEN, self.dcim.oob_url(self.oob_info)])


class OobError(Exception):
    pass



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime
import json
import urllib

from bmcmanager.logs import log
from bmcmanager.firmwares.base import LatestFirmwareFetcher


TRACKED_FIRMWARE = {
    'ThinkServer System Manager (TSM) Update Bundle File - ThinkServer Systems': 'TSM',  # noqa
    'Firmware for Avago 9340-8i/ 9364-8i/ 9380-8e SAS3 RAID and RAID 720i/720ix AnyRAID Adapter - ThinkServer Systems': 'RAID',  # noqa
    'Firmware for Avago 9240-8i SAS RAID and ThinkServer 510i AnyRAID Adapter  - ThinkServer Systems': 'RAID',  # noqa
    'Firmware for LSI 9270CV-8i/ 9286CV-8e SAS RAID Card for Windows and Linux - ThinkServer Systems': 'RAID',  # noqa
    'Firmware Update for Delta DPS-750AB-21A PT750W Power Supply Unit (PSU) - ThinkServer RD350, RD450, RD550, RD650, TD350': 'PSU',  # noqa
    'Firmware Update for Delta DPS-750AB-20A TT750W Power Supply Unit (PSU) - ThinkServer RD350, RD450, RD550, RD650, TD350': 'PSU',  # noqa
    'Firmware Update for Liteon PS-2751-3L PT750W Power Supply Unit (PSU) - ThinkServer RD350, RD450, RD550, RD650, TD350': 'PSU',  # noqa
    'Firmware Update for Liteon PS-2551-6L-LF PT550W Power Supply Unit (PSU) - ThinkServer RD350, RD450, RD550, RD650, TD350': 'PSU',  # noqa
    'Firmware Update for Delta DPS-550AB-5A PT550W Power Supply Unit (PSU) - ThinkServer RD350, RD450, RD550, RD650, TD350': 'PSU',  # noqa
    'Firmware Update for Delta DPS-1100EBA PT1100W Power Supply Unit (PSU) - ThinkServer  RD450, RD550, RD650, TD350': 'PSU',  # noqa
}


LENOVO_URL = 'https://datacentersupport.lenovo.com/gr/en/api/v4/downloads/drivers?productId=servers/{}/{}'  # noqa


class LenovoBase(LatestFirmwareFetcher):
    model_name = 'model-name'       # replace in sub-classes, e.g. 'thinkserver'
    device_name = 'device-name'     # replace in sub-classes, e.g. 'rd550'
    extra_firmware = {
        # add here any extra firmware items to track
    }

    def get(self):
        url = LENOVO_URL.format(self.model_name, self.device_name)
        try:
            log.debug('GET {}'.format(url))
            response = json.loads(urllib.request.urlopen(url).read())
            items = response['body']['DownloadItems']
        except (json.JSONDecodeError, urllib.error.URLError) as e:
            log.error('Could not fetch URL: {}'.format(e))
            return {}, []
        except KeyError as e:
            log.error('Invalid data format: {}'.format(e))
            return {}, []

        tracked_firmware = {**TRACKED_FIRMWARE, **self.extra_firmware}

        result = []
        downloads = []
        for item in items:
            try:
                component = tracked_firmware.get(item['Title'])
                if component is None:
                    continue

                fw = next(filter(
                    lambda f: f['TypeString'] != 'TXT README', item['Files']))

                downloads.append(fw['URL'])

                timestamp = datetime.fromtimestamp(fw['Date']['Unix'] / 1000)
                result.append({
                    'component': component,
                    'name': item['Title'],
                    'version': item['SummaryInfo']['Version'],
                    'date': str(timestamp),
                    'file': fw['URL'],
                })

            except ValueError as e:
                log.error('Invalid data format: {}'.format(e))

            except KeyError as e:
                log.error('Missing information: {}'.format(e))

        return result, downloads


class RD550(LenovoBase):
    model_name = 'thinkserver'
    device_name = 'rd550'
    extra_firmware = {
        'BIOS Update Bundle File - ThinkServer RD550': 'BIOS',
    }


class RD350(LenovoBase):
    model_name = 'thinkserver'
    device_name = 'rd350'
    extra_firmware = {
        'ThinkServer BIOS Update Bundle File - ThinkServer RD350, RD450': 'BIOS',  # noqa
    }



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from bmcmanager.commands.base import BMCManagerServerCommand, BMCManagerServerListCommand


class Get(BMCManagerServerListCommand):
    """
    print system event logs
    """
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--analysed', action='store_true', default=False,
            help='decode OEM specific fields')
        return parser

    def action(self, oob):
        if self.parsed_args.analysed:
            return oob.ipmi_logs_analysed()
        else:
            return oob.ipmi_logs()


class Clear(BMCManagerServerCommand):
    """
    clear system event logs
    """
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--from-firmware-upgrades', action='store_true', default=False,
            help='clear logs from firmware upgrades only')
        return parser

    def action(self, oob):
        if self.parsed_args.from_firmware_upgrades:
            oob.clear_firmware_upgrade_logs()
        else:
            oob.clear_ipmi_logs()



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


class DcimBase(object):
    """
    Base DCIM class
    """
    def __init__(self, args, config):
        self.args = args
        self.config = config
        self.identifier = args.server
        self.dcim_params = config
        self.api_url = config['api_url']
        self.is_rack = self.args.type == 'rack'
        self.is_rack_unit = self.args.type == 'rack-unit'
        self.is_serial = self.args.type == 'serial'

    def get_info(self):
        raise NotImplementedError('get_info not implemented')

    def get_oobs(self):
        raise NotImplementedError('get_oobs not implemented')

    def oob_url(self):
        raise NotImplementedError

    def supports_secrets(self):
        return getattr(self, 'get_secret') is not None

    def set_custom_fields(self, oob_info, custom_fields):
        return False


class DcimError(Exception):
    pass



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re

from bmcmanager import nagios


def version_tuple(version_str):
    """converts '1.2.3' --> (1, 2, 3)."""
    return tuple(int(x) for x in version_str.split('.'))


def check_version_strings(version_value, expected_version, version_check=True):
    try:
        version = version_tuple(version_value)

    except (TypeError, ValueError, AttributeError):
        return nagios.CRITICAL, 'invalid data'

    try:
        target_version = version_tuple(expected_version)

    except (TypeError, ValueError, AttributeError):
        return nagios.CRITICAL, 'missing config'

    if version >= target_version:
        return nagios.OK, 'ok: {}'.format(version_value)
    else:
        msg = 'have {}, expected {}'.format(version_value, expected_version)
        if version[0] < target_version[0]:
            status = nagios.CRITICAL
        else:
            status = nagios.WARNING

        if not version_check:
            status = nagios.OK
        return status, msg


def psu_checks(all_psu_versions, version_dict):
    try:
        psu_info = all_psu_versions.split(', ')
    except (ValueError, TypeError, AttributeError):
        return {
            'psu': (nagios.CRITICAL, 'invalid data')
        }

    results = {}
    for psu in psu_info:
        m = re.match(
            r'^(?P<psu_slot>\d+)\/(?P<psu_type>\w*)'
            r':\s(?P<psu_version>\d+\.\d+\.\d+)$',
            psu)

        if not m or not m.groupdict():
            results['psu'] = nagios.CRITICAL, 'invalid data'
            continue

        psu_info_dict = m.groupdict()
        psu_slot = psu_info_dict['psu_slot']
        psu_type = psu_info_dict['psu_type']

        version_str = psu_info_dict['psu_version']
        target_version_str = version_dict.get('psu_{}'.format(
            psu_type.lower()))
        res, msg = check_version_strings(
            version_str, target_version_str, version_check=False)

        results['psu-{}'.format(psu_slot)] = res, '{}, {}'.format(
            msg, psu_type)

    return results


def check_firmware(custom_fields, oob_params):
    result, msg = nagios.OK, []

    psus = psu_checks(custom_fields['PSU'], oob_params['oob_params'])
    checks = {
        'bios': check_version_strings(
            custom_fields['BIOS'], oob_params['oob_params'].get('bios')),
        'tsm': check_version_strings(
            custom_fields['TSM'], oob_params['oob_params'].get('tsm')),
        **psus,
    }

    for check, check_data in checks.items():
        state, text = check_data
        msg.append('{} ({})'.format(check.upper(), text))
        result = max(state, result)

    try:
        expected_psus = int(oob_params['oob_params']['expected_psus'])
        if expected_psus != len(psus):
            msg.append('{} PSUs present (expected {})'.format(
                len(psus), expected_psus))
            result = max(state, nagios.CRITICAL)

    except (TypeError, ValueError, KeyError):
        pass

    return result, msg
